import {Net} from "./Net";
import {DefaultConstructor, Duration, Json, Minute, Second} from "mvx-utils";
import {Method} from "./http/HttpBuilder";
import {SignNonceStr} from "../Sign";
import {Token} from "../model/Token";

export enum Code {
  Success = 200,
  TokenExpire = 401,
  ElseError = 500,
}

export class CodeError extends Error{
  constructor(err:string|Error, public code:Code = Code.ElseError) {
    super();
    this.name = "CodeError";
    if (typeof err === 'string') {
      this.message = `${err}`
    } else {
      this.message = `${err.message}`;
      this.stack = err.stack
    }
  }
}

export interface Decoder {
  decode<T extends object|string>(uri:string, resType:DefaultConstructor<T>, res:string, err:CodeError) : [T, CodeError|null];
}

export interface Encoder {
  encode<T extends object>(request:object) : [string, Error|null];
}

interface opt {
  encoder:Encoder;
  header:Map<string, string>;
  setToken:(opt:{header:Map<string, string>, request:object}, token:string)=>void;
  method:Method;
  cacheTTL:(codeError:Code) => Duration;
  cacheAble:boolean;
  decoder:Decoder;
}

export type HttpOption = (o:opt)=>Error|null
function newHttpOpt(...opts:HttpOption[]):opt {
  let opt:opt = {
    header:new Map<string, string>(),
    encoder:new class implements Encoder {
      encode<T extends object>(request: object): [string, (Error | null)] {
        throw new Error(`encoder not found, please call WithJsonBody、WithFormUrlEncodedBody、WithBody`);
      }
    },
    method:"GET",
    setToken:(opt, token) => {},
    cacheTTL:(code) => {return 0;},
    cacheAble:false,
    decoder:new class implements Decoder {
      decode<T extends object|string>(uri:string, resType: DefaultConstructor<T>, res: string, err: CodeError): [T, (CodeError | null)] {
        const retStr = Object.prototype.toString.call(new resType()) == "[object String]";
        console.log("decoder.decode", "uri", uri, "res", res, "retStr", retStr);

        if (retStr) {
          return [<T>res, err];
        }

        if (err.code!==Code.Success) {
          return [new resType(), err]
        }
        if (res === "") {
          return [new resType(), new CodeError("response is empty")]
        }

        let [resT, err1] = new Json().fromJson(res, <DefaultConstructor<object>>resType);
        if (err1) {
          return [new resType(), new CodeError(err1)]
        }

        return [<T>resT, null]
      }
    },
  };

  opts.forEach(f=> {
    f(opt);
  });

  return opt;
}

export function WithHeader(header:Map<string, string>):HttpOption {
  return (o:opt):Error|null => {
    header.forEach((value, key) => {
      o.header.set(key, value);
    });
    return null;
  }
}

export function WithBearerToken():HttpOption {
  return (o:opt):Error|null => {
    o.setToken = (opt, token) => {
        o.header.set("Authorization", `Bearer ${token}`);
    };
    return null;
  }
}

export function WithBodyToken(tokenKey:string = "token"):HttpOption {
  return (o:opt):Error|null => {
    o.setToken = (opt, token) => {
      (<Record<string, string>>opt.request)[tokenKey] = token;
    };
    return null;
  }
}

export function WithoutToken():HttpOption {
  return (o:opt):Error|null => {
    o.setToken = (opt, token) => {};
    return null;
  }
}

export function WithReqId():HttpOption {
  return (o:opt):Error|null => {
    o.header.set("X-Req-Id", SignNonceStr());
    return null;
  }
}

export function WithBody(encoder:Encoder):HttpOption {
  return (o:opt):Error|null => {
    o.encoder = encoder;
    return null;
  }
}

export function WithFormUrlEncodedBody():HttpOption {
  return (o:opt):Error|null => {
    o.header.set("X-Requested-With", "XMLHttpRequest");
    o.header.set('Content-Type', 'application/x-www-form-urlencoded');

    o.encoder =  new class implements Encoder {
      encode<T extends object>(request: object): [string, (Error | null)] {
        const [json, err] = new Json().toJson(request);
        if (err) {
          return ["", err];
        }

        const args = JSON.parse(json);
        const ret = Object.keys(args).map((key) => {
          if ((<any>args)[key] == null) {
            return encodeURIComponent(key) + "="
          }

          return encodeURIComponent(key) + "=" + encodeURIComponent((<any>args)[key]);
        }).join("&");
        return [ret, null];
      }
    };

    return null;
  }
}

export function WithJsonBody():HttpOption {
  return (o:opt):Error|null => {
    o.header.set("X-Requested-With", "XMLHttpRequest");
    o.header.set('Content-Type', 'application/json;charset=utf8');
    o.encoder = new class implements Encoder {
      encode<T extends object>(request: object): [string, (Error | null)] {
        return new Json().toJson(request);
      }
    };

    return null;
  }
}

export function WithHttpMethod(method:Method):HttpOption {
  return (o:opt):Error|null => {
    o.method = method;
    return null;
  }
}

export function WithCache(cacheAble:boolean, cacheTTL:(code:Code) => Duration = (code => code===Code.Success ? 1*Minute : 5*Second)):HttpOption {
  return (o:opt):Error|null => {
    o.cacheTTL = cacheTTL;
    o.cacheAble = cacheAble;
    return null;
  }
}

export function WithHttpDecoder(decoder:Decoder):HttpOption {
  return (o:opt):Error|null => {
    o.decoder = decoder;
    return null;
  }
}

const cache:Map<string, {value:[string, CodeError], expireTime:Duration}> = new Map();
const waiting:Map<String, {resolve:((ret:[any, CodeError])=>void), opt:opt, resType:DefaultConstructor<any>}[]> = new Map();
export async function SendNoToken<T extends object|string>(uri:string,
                                             request:object,
                                             resType:DefaultConstructor<T>,
                                             net:Net, ...opts:HttpOption[]):Promise<[T, CodeError|null]> {
  // 兼容uri带有参数情况
  if (uri.lastIndexOf("?")>=0) {
    const parts = uri.substring(uri.lastIndexOf("?") + 1).split("&");
    parts.forEach(r => {
      const args = r.split('=');
      if(args && args.length>=2) {
        (<any>request)[args[0]] = args[1];
      }
    });
    uri = uri.substring(0, uri.lastIndexOf("?"));
  }

  const opt = newHttpOpt(...opts);
  opt.setToken({header:opt.header, request:request}, await net.getToken());

  const [body, err0] = opt.encoder.encode(request);
  if (err0) {
    return [new resType(), new CodeError(err0.message)]
  }

  const builder = net.getHttpBuilder();
  builder.setUri(uri).setMethod(opt.method).setHeaders(opt.header).setContent(body);

  const reqId = `${builder.method()}@${builder.getUrl()}/${builder.content()}`;
  if (opt.cacheAble && cache.has(reqId)) {
    console.log(`[${builder.method()}] ${builder.getUrl()} use cache req `, builder.content());
    const response = cache.get(reqId)!;
    const [res, err] = response.value;
    console.log(`[${builder.method()}]  ${builder.getUrl()} use cache res `, response);
    return opt.decoder.decode(uri, resType, res, err);
  }

  console.log(`[${builder.method()}] ${builder.getUrl()} send req `, builder.content());

  if (!waiting.has(reqId)) {
    waiting.set(reqId, []);
  }

  const promise = new Promise<[T, CodeError|null]>(resolve => {
    waiting.get(reqId)!.push({resolve:resolve, opt:opt, resType:resType});
  });

  if (waiting.get(reqId)!.length>1) {
    return promise;
  }

  const [res, err] = await builder.build().send();
  console.log(`[${builder.method()}]  ${builder.getUrl()} send request `, [res, err]);

  if (opt.cacheAble) {
    const cacheTTL = opt.cacheTTL(err.code);
    cache.set(reqId, {value:[res, err], expireTime:new Date().getTime()+cacheTTL});
  }

  waiting.get(reqId)!.forEach(r => {
    const ret = <[T, CodeError|null]>r.opt.decoder.decode(uri, r.resType, res, err);
    r.resolve(<any>ret);
    console.log(`[${builder.method()}]  ${builder.getUrl()} decode response `, ret);
  });
  waiting.set(reqId, []);

  return promise;
}

export async function Send<T extends object|string>(uri:string,
                                      request:object,
                                      resType:DefaultConstructor<T>,
                                      net:Net, ...opts:HttpOption[]):Promise<[T, CodeError|null]> {
  if (net.is401()) {
    return [new resType(), new CodeError("has 401", Code.TokenExpire)]
  }

  let token = await net.getToken();
  // token为空，需要处理401，并且不执行真正的网络请求。
  if (token === Token.empty) {
    await net.process401();
    return [new resType(), new CodeError("token is empty", Code.TokenExpire)]
  }

  let [ret, err] = await SendNoToken(uri, request, resType, net, ...opts);
  if (!err || err.code !== Code.TokenExpire) {
    // 取消401
    net.clear401();
    return [ret, err]
  }

  // 返回401时，如果请求的token与现在存储token不是一样的，说明有登录接口修改过，最新
  // 存储的token是否过期，无法判断，所以不能执行真正的401操作
  // 此处不能通过net层is401来判断，还没有开始处理401或者其他接口处理完401的情况下，is401都是false
  // 但是后一种情况却不能执行401；对token而言后一种情况token一定不一样，却可以分辨出两种情况
  if (token !== await net.getToken()) {
    return [new resType(), new CodeError("token is too old", Code.TokenExpire)]
  }

  await net.process401();
  return [new resType(), new CodeError("token is expire", Code.TokenExpire)]
}